/****(c) 2020. A. Gerbessiotis. See ../KAagcopy.txt ****/
/*******************************************************/
/*
#include <math.h>
*/
#include <stdio.h>
#include <sys/time.h>
#include "agsmrgd.h"

#define ALL_FUNCTIONS
#undef  ALL_FUNCTIONS
#undef  FEW_FUNCTIONS
#define FEW_FUNCTIONS
#define NALGS 7    /* Make sure there are as many entries in gfunc/doalg below or CRASH!!!! */

#define ELMNTS 2
#define MODULO 100000
#define MODULO_UINT ((unsigned int) 0x7FFFFFFF)
#define AVG_TINIT   AVG_PERRUN = AVG_ELAPSED =(double)0;
#define AVG_TSTART  AVG_PERRUN = (double) 0.0 ; avg_tstart();
#define AVG_TEND    avg_tend(); AVG_PERRUN = (avg_treturn() ); AVG_ELAPSED +=(AVG_PERRUN);
#define AVG_PERRUN  per_run
#define AVG_ELAPSED elapsed

static struct timeval avgtstart, avgtend;
static struct timezone avgtstruct;

void   avg_tstart  (void);
void   avg_tend    (void);
double avg_treturn (void);
void   printseq    (data *inp,int n,char *s,int debug);
void   dataseq     (data *d, int n, int type);
void   datamcheck   (data *res1, data *res2, data *ori , int n);
void   seqdatacheck(data *ori , int n);

void   (*ITYPEMERG)(data *,data*,int,int);


void avg_tstart(void) {
  gettimeofday(&avgtstart, &avgtstruct);
}

void avg_tend(void) {
  gettimeofday(&avgtend,&avgtstruct);
}

double avg_treturn(void) {
  double t1,t2;
  t1=(double)avgtstart.tv_sec + (double)avgtstart.tv_usec*(0.000001);
  t2=(double)  avgtend.tv_sec + (double)  avgtend.tv_usec*(0.000001);
  return(t2-t1);
}

int g1compare(const void *x,const void *y)
{
	return(int)(* ((data*)x)-*((data*)y));
}

int main(int argc,char *argv[]) {
  int j,k,n,runs, type, debug;
  data *ptr1, *ptr2, *bptr1, *bptr2, *sptr;
  double etime;

  void (*gfunc[NALGS])() = {  
    d8mf1    ,   d8mf2   ,  d8mf3   ,  d8mf4   ,  d8mf4o,  
    d8mw3   ,   d8mw4
   };

  char *labels[NALGS]   = { 
    " d8mf1" ,  " d8mf2" , " d8mf3" , " d8mf4" , "d8mf4o" , 
    " d8mw3" , " d8mw4"
   };

  int doalg[NALGS] = { 
    1,1,1,1,1,   1,1
   };

   ptr1=bptr1=sptr=ptr2=bptr2=NULL;

   n=1000;   runs=1;
   if (( argc <= 1 ) || (argc > 5)) {
             printf("Usage: %s n type runs debug\n",argv[0]);
             printf("Type : 1=rand MODULO_UINT, 2=rand (INT_MAX), 3= INC, 4= 17  else= DEC\n");
             return(1);
   } 
     else {
             n    = atoi(argv[1]);  /*  size  */
             type = atoi(argv[2]);  /*  type  */
             runs = atoi(argv[3]);  /*  runs  */
             debug= atoi(argv[4]);  /*  debug */
             printf("%d : %d %d %d %d\n",argc,n,type,runs,debug);
   }
   etime=0.0;

   sptr= (data *) malloc(2*n*sizeof(data));
   bptr1= (data *) malloc(n*sizeof(data));
   bptr2= (data *) malloc(n*sizeof(data));
   ptr1 = (data *) malloc(n*sizeof(data));
   ptr2 = (data *) malloc(n*sizeof(data));

   memset(sptr,0,2*n*sizeof(data));
   memset(bptr1,0,n*sizeof(data));
   memset(bptr2,0,n*sizeof(data));
   memset( ptr1,0,n*sizeof(data));
   memset( ptr2,0,n*sizeof(data));


   srandom(21+1001*11);
   dataseq (bptr1,n,type);                     /* source for all runs :1 */
   qsort   (bptr1,n,sizeof(data),g1compare);
   printseq(bptr1,n,"In1",debug);
   memcpy  (ptr1,bptr1,n*sizeof(data));        /* for merging : 1 */

   dataseq (bptr2,n,type);                     /* source for all runs :2 */
   qsort   (bptr2,n,sizeof(data),g1compare);
   printseq(bptr2,n,"In2",debug);
   memcpy  (ptr2,bptr2,n*sizeof(data));        /* for merging : 2 */

   memcpy(sptr,bptr1,n*sizeof(data));
   memcpy(&sptr[n],bptr2,n*sizeof(data));
     etime=0.0;
     avg_tstart();
   qsort(sptr,2*n,sizeof(data),g1compare);     /* sort union of 1 and 2  */
     avg_tend();
     etime += (avg_treturn());
         printf("OK:%10d:%1d:%2d:%1d:",n,type,1,debug);
         printf(   "%11s : %12.8f : %12.3f\n","qsort",etime,etime);
         fflush(stdout);
   printseq( sptr   , n, "O1a", debug);
   printseq(&sptr[n], n, "O1b", debug);

   for(k=0;k<NALGS;k++) {
      if (doalg[k]>=1) {
         ITYPEMERG=gfunc[k];
         etime=0.0;

         for(j=0;j<runs;j++) {
            memcpy(ptr1,bptr1,n*sizeof(data));
            memcpy(ptr2,bptr2,n*sizeof(data));
            avg_tstart();
               ITYPEMERG(ptr1,ptr2,n,n);
            avg_tend();
            etime += (avg_treturn());
         }
/*       memcmp 3 lines below does not tell you which keys are out of sequence
*/
         datamcheck(ptr1,ptr2,sptr,n);

      if ((memcmp(ptr1,sptr,n*sizeof(data))  != 0) || (memcmp(ptr2,&sptr[n],n*sizeof(data)) !=0 )){
         printf("NO:%10d:%1d:%2d:%1d:",n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n",labels[k],etime,etime/runs);
      }
      else {
         printf("%10d:%1d:%2d:%1d:",n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n",labels[k],etime,etime/runs);
         fflush(stdout);
         printseq( ptr1   , n, "O2a", debug);
         printseq( ptr2   , n, "O2b", debug);
      }
      }
   }

   free((void *) ptr2);
   free((void *) ptr1);
   free((void *) bptr2);
   free((void *) bptr1);
   free((void *) sptr);
   return(0);
}

void printseq(data *inp, int n, char *s, int debug) 
{
 register int i;
 if (1==debug)  { 
 fprintf(stdout,"[%4s]: ",s);
 fflush(stdout);
 for(i=0;i<ELMNTS;i++){
   fprintf(stdout,"%10d,", (data) inp[i]);
 }
/*
   fprintf(stdout,"...\n          ");
*/
   fprintf(stdout,"... ");
 for(i=n-ELMNTS;i<n;i++){
   fprintf(stdout,"%10d,", (data) inp[i]);
 }
 fprintf(stdout,"\n");
 fflush(stdout);
 }
}

void dataseq  (data *d, int n, int type)
{
 register int i;
/* Commented out for merging
    srandom(21+1001*11);
*/
    if (type == 1) {
        for (i=0;i<n;i++) {
           d[i]= (data ) (random() & MODULO_UINT);
        }
     } 
     else 
     if (type == 2) {
         for (i=0;i<n;i++) {
            d[i]= (data )  ((n)*(random()/((double)INT_MAX)));
         }
     }
     else
     if (type == 3) {
         for (i=0;i<n;i++) {
            d[i]= (data )  (i+1);
         }
     }
     else
     if (type == 4) {
         for (i=0;i<n;i++) {
            d[i]= (data )  (17);
         }
     }
     else {
       for (i=0;i<n;i++) {
         d[i]= (data )  (n-i);
       }
     }
}



void seqdatacheck (data *ori, int n)
{
  register int i;
   for(i=0;i<n-1;i++) {
     if (ori[i] > ori[i+1]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,ori[i],ori[i+1]);
      fflush(stdout); 
      return;
     }
   }
   fprintf(stdout,"OK:"); fflush(stdout);
}

void datamcheck (data *res1, data *res2, data *ori , int n)
{
  register int i;
   for(i=0;i<n;i++) {
     if (res1[i] != ori[i]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,res1[i],ori[i]);
      fflush(stdout);
      return;
     }
   }
   for(i=0;i<n;i++) {
     if (res2[i] != ori[n+i]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,res2[i],ori[n+i]);
      fflush(stdout);
      return;
     }
   }
   fprintf(stdout,"OK:");fflush(stdout);
}

