/****(c) 2020. A. Gerbessiotis. See ../KAagcopy.txt ****/
/*******************************************************/
/*
#include <math.h>
*/
#include <stdio.h>
#include <sys/time.h>
#include "srtseq/agssrtd.h"

#define ALL_FUNCTIONS
#undef  ALL_FUNCTIONS
#undef  FEW_FUNCTIONS
#define FEW_FUNCTIONS
#define NALGS 27   /* Make sure # entries in gfunc/doalg is NALGS or CRASH!!!! */

#define ELMNTS 8
#define MODULO 100000
#define MODULO_UINT ((unsigned int) 0x7FFFFFFF)
#define AVG_TINIT   AVG_PERRUN = AVG_ELAPSED =(double)0;
#define AVG_TSTART  AVG_PERRUN = (double) 0.0 ; avg_tstart();
#define AVG_TEND    avg_tend(); AVG_PERRUN = (avg_treturn() ); AVG_ELAPSED +=(AVG_PERRUN);
#define AVG_PERRUN  per_run
#define AVG_ELAPSED elapsed

static struct timeval avgtstart, avgtend;
static struct timezone avgtstruct;

void   avg_tstart  (void);
void   avg_tend    (void);
double avg_treturn (void);
void   dprintseq    (ddata *inp,int n,char *s,int debug);
void   ddataseq     (ddata *d, int n, int type);
void   ddatacheck   (ddata *res, ddata *ori , int n);
void   dseqdatacheck(ddata *ori , int n);

void   (*DTYPESORT)(ddata *,int);


void avg_tstart(void) {
  gettimeofday(&avgtstart, &avgtstruct);
}

void avg_tend(void) {
  gettimeofday(&avgtend,&avgtstruct);
}

double avg_treturn(void) {
  double t1,t2;
  t1=(double)avgtstart.tv_sec + (double)avgtstart.tv_usec*(0.000001);
  t2=(double)  avgtend.tv_sec + (double)  avgtend.tv_usec*(0.000001);
  return(t2-t1);
}

int g1compare(const void *x,const void *y)
{
	return(int)(* ((ddata*)x)-*((ddata*)y));
}

int main(int argc,char *argv[]) {
  int j,k,n,runs, type, debug;
  ddata *ptr,*bptr, *sptr;
  double etime;

  void (*ifunc[NALGS])() = {  
   d1ins1,   d1ins2,    d1ins3,   d1ins4,   d1ins5,
   d1ins6,   d1ins7,    d1ins8,   d1inb1,   d4mrgi1, 
   d4mrgi2,  d3mrgiu,   d3mrgio,  d5hpx1,   d5hpx3, 
   d6carh,   d6clrs,    d6qkrd1,  d6qkrf1,  d6qkrw1,
   d6qkrw3,  d9rdx4u,   d7qknd1,  d7qknd2,  d9rdx4u,
   d9rdx4o,  d6qkrw5
   };

  char *labels[NALGS]   = { 
   "d1ins1",   "d1ins2",    "d1ins3",   "d1ins4",   "d1ins5",
   "d1ins6",   "d1ins7",    "d1ins8",   "d1inb1",   "d4mrgi1", 
   "d4mrgi2",  "d3mrgiu",   "d3mrgio",  "d5hpx1",   "d5hpx3", 
   "d6carh",   "d6clrs",    "d6qkrd1",  "d6qkrf1",  "d6qkrw1",
   "d6qkrw3",  "d9rdx4u",   "d7qknd1",  "d7qknd2",  "d9rdx4u",
   "d9rdx4o",  "d6qkrw5"
   };

  int doalg[NALGS] = { 
   1,1,1,1,1,   1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
   1,1
/*
   0,0,0,0,0,   0,0,0,0,0,  0,0,0,0,0,  1,1,1,1,1,  0,0,0,0,0,
   0,0
   0,0,0,0,0,   0,0,0,0,1,  1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
   1,1
*/
   };






   ptr=bptr=sptr=NULL;

   n=1000;   runs=1;
   if (( argc <= 1 ) || (argc > 5)) {
             printf("Usage: %s n type runs debug\n",argv[0]);
             printf("Type : 1=rand MODULO_UINT, 2=rand (INT_MAX),");
             printf(" 3= INC, 4= 17  else= DEC\n");
             printf("Quick: BSRTSZE=%d INQSSZE=%d SPLITTER=%d\n",
                     BSRTSZE,INQSSZE,SPLITTER); 
             printf("Split:  0:left, 1:middle, 2:random\n");
             return(1);
   } 
     else {
             n    = atoi(argv[1]);  /*  size  */
             type = atoi(argv[2]);  /*  type  */
             runs = atoi(argv[3]);  /*  runs  */
             debug= atoi(argv[4]);  /*  debug */
             printf("%d : %d %d %d %d\n",argc,n,type,runs,debug);
   }
   etime=0.0;
   sptr= (ddata *) malloc(n*sizeof(ddata));
   bptr= (ddata *) malloc(n*sizeof(ddata));
   ptr = (ddata *) malloc(n*sizeof(ddata));
   memset( ptr,0,n*sizeof(ddata));
   memset(bptr,0,n*sizeof(ddata));
   memset(sptr,0,n*sizeof(ddata));

   ddataseq(bptr,n,type);
   dprintseq(bptr,n,"InQ",debug);
   memcpy(ptr,bptr,n*sizeof(ddata));

   qsort(ptr,n,sizeof(ddata),g1compare);
   dprintseq( ptr,n,"OtQ",debug);
   memcpy(sptr,ptr,n*sizeof(ddata));


  etime=0.0;
  for(j=0;j<runs;j++){
       memcpy(ptr,bptr,n*sizeof(ddata));
       avg_tstart();
          qsort(ptr,n,sizeof(ddata),g1compare);
       avg_tend();
       etime += (avg_treturn());
  }
       ddatacheck(ptr,sptr,n);

      if (memcmp(ptr,sptr,n*sizeof(ddata))  != 0){
         printf("NO:%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
               BSRTSZE,INQSSZE,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n","qsort",
               etime,etime/runs);
      }
      else {
         printf("%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
               BSRTSZE,INQSSZE,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n","qsort",
               etime,etime/runs);
         fflush(stdout);
      }
  etime=0.0;
  for(j=0;j<runs;j++){
       memcpy(ptr,bptr,n*sizeof(ddata));
       avg_tstart();
          i0gnu(ptr,n,sizeof(ddata),g1compare);
       avg_tend();
       etime += (avg_treturn());
  }
       ddatacheck(ptr,sptr,n);

      if (memcmp(ptr,sptr,n*sizeof(ddata))  != 0){
         printf("NO:%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
                 BSRTSZE,INQSSZE,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n","i0gnu",etime,etime/runs);
      }
      else {
         printf("%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
                 BSRTSZE,INQSSZE,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n","i0gnu",etime,etime/runs);
         fflush(stdout);
      }

   for(k=0;k<NALGS;k++) {
      if (doalg[k]>=1) {
         DTYPESORT=ifunc[k];
         etime=0.0;

         for(j=0;j<runs;j++) {
            memcpy(ptr,bptr,n*sizeof(ddata));
            avg_tstart();
               DTYPESORT(ptr,n);
            avg_tend();
            etime += (avg_treturn());
         }
         /*memcmp below does not tell which keys are out of order*/
         ddatacheck(ptr,sptr,n);

      if (memcmp(ptr,sptr,n*sizeof(ddata))  != 0){
         printf("NO:%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
                 BSRTSZE,INQSSZE,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n",labels[k],etime,etime/runs);
      }
      else {
         printf("%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
                 BSRTSZE,INQSSZE,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n",labels[k],etime,etime/runs);
         fflush(stdout);
      }
      }
   }

   free((void *) ptr);
   free((void *) bptr);
   free((void *) sptr);
   return(0);
}

void dprintseq(ddata *inp, int n, char *s, int debug) 
{
 register int i;
 if (1==debug)  { 
 fprintf(stdout,"[%4s]: ",s);
 fflush(stdout);
 for(i=0;i<ELMNTS;i++){
   fprintf(stdout,"%10d,", (ddata) inp[i]);
 }
   fprintf(stdout,"...\n          ");
 for(i=n-ELMNTS;i<n;i++){
   fprintf(stdout,"%10d,", (ddata) inp[i]);
 }
 fprintf(stdout,"\n");
 fflush(stdout);
 }
}

void ddataseq  (ddata *d, int n, int type)
{
 register int i;
    srandom(21+1001*11);
    if (type == 1) {
        for (i=0;i<n;i++) {
           d[i]= (ddata ) (random() & MODULO_UINT);
        }
     } 
     else 
     if (type == 2) {
         for (i=0;i<n;i++) {
            d[i]= (ddata )  ((n)*(random()/((double)INT_MAX)));
         }
     }
     else
     if (type == 3) {
         for (i=0;i<n;i++) {
            d[i]= (ddata )  (i+1);
         }
     }
     else
     if (type == 4) {
         for (i=0;i<n;i++) {
            d[i]= (ddata )  (17);
         }
     }
     else {
       for (i=0;i<n;i++) {
         d[i]= (ddata )  (n-i);
       }
     }
}



void dseqdatacheck (ddata *ori, int n)
{
  register int i;
   for(i=0;i<n-1;i++) {
     if (ori[i] > ori[i+1]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,ori[i],ori[i+1]);
      fflush(stdout); 
      return;
     }
   }
   fprintf(stdout,"OK:"); fflush(stdout);
}

void ddatacheck (ddata *res, ddata *ori , int n)
{
  register int i;
   for(i=0;i<n;i++) {
     if (res[i] != ori[i]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,res[i],ori[i]);
      fflush(stdout);
      return;
     }
   }
   fprintf(stdout,"OK:");fflush(stdout);
}
