/****Copyright (c) 1997-2014 All rights reserved
 ****A. V. Gerbessiotis 
 ****Permission to use, copy, modify, and distribute this software,
 ****and to incorporate it, in whole or in part, into 
 ****other software,is hereby granted without fee, provided that
 ****  (1) the above copyright notice and this permission 
 ****      notice appear in all copies of the source code, and 
 ****      the above copyright notice appear in clearly visible 
 ****      form on all supporting documentation and distribution 
 ****      media, and
 ****     (2) any redistribution of the software, in original
 ****         or modified form, be without fee and subject to
 ****         these same conditions.
 ****   No guarantee, whatsoever, is offered that the code works as
 ****   advertised or is absent of any, even damaging,
 ****   side-effects. Use this code at your own personal risk.
 ****/

#include "issrt.h"
#include "avgimain.h"

#define N_SORT_ALGS 19
#define DEBUG 0

int compare(const void *x,const void *y)
{
	return(int)(* ((idata*)x)-*((idata*)y));
}

int main(int argc,char *argv[]) {
  int i,j,k,n,runs, type;
  idata *ptr,*bptr, *obptr;
  double elapsed;


/* BEWARE of ignu01: it is [12] ie 13th function in sequence */

  void (*gfunc[N_SORT_ALGS])() = {  
 i32rdx, i32rdx2, i32rdx4, o32rdx4, 
 i21ins, i22ins,  i01ins, 
 i30mi, i34mi, i43hx, i44hx, i45hx,
 ignu01, i50qw, i52dbh, i63fbc, i55non0, i55non1, o50qw
};

  char *labels[N_SORT_ALGS]   = { 
"i32rdx", "i32rdx2", "i32rdx4", "o32rdx4", 
"i21ins", "i22ins",  "i01ins" , 
"i30mi",  "i34mi",   "i43hx", "i44hx", "i45hx",
"ignu01", "i50qw",   "i52dbh",  "i63fbc", "i55non0", "i55non1", "o50qw"
  };

  int doalg[N_SORT_ALGS] = { 
1,1,1,1,
0,0,0,
0,0,0,0,0,
0,0,0,0,0,0,1
/*
1,1,1,1,
0,0,0,
1,1,1,1,1,
1,1,1,1,1,1,1
*/
  };

  void (*GTYPESORT)(idata *,int);

  n=1000;runs=1;
  if (( argc <= 1 ) || (argc > 4)) {
    printf("Usage: %s size runs type\nDEFAULTS= THRESI16=%d TMERGEI=%d SPLITTER=%d (0:left,1:middle,2:random)\nType 1=rand MODULO_UINT, 2=rand (INT_MAX), 3= 10, 4=i 5 =n-i+1\n",argv[0],THRESI16,TMERGEI,SPLITTER);
    return(1);
  } 
  else {
    n = atoi(argv[1]);
    runs=atoi(argv[2]);
    type=atoi(argv[3]);
   }

/*
   ptr = (idata*) Malloc(n*sizeof(idata),"ptr ");
   bptr= (idata*) Malloc(n*sizeof(idata),"bptr ");
*/
   ptr = (idata*) malloc(n*sizeof(idata));
   bptr= (idata*) malloc(n*sizeof(idata));
   obptr= (idata*) malloc(n*sizeof(idata));
   memset( ptr,0,n*sizeof(idata));
   memset(bptr,0,n*sizeof(idata));
   memset(obptr,0,n*sizeof(idata));

   srandom(21+1001*11);
   for(i=0;i<n;i++)  {
    if  ( type==1 ) {
     bptr[i]=ptr[i]=(idata) (random()&MODULO_UINT);
    }
    if  ( type==2 ) {
     bptr[i]=ptr[i]=(idata) ((n)*(random()/((double)INT_MAX)));
    }
    if  ( type==3 ) {
     bptr[i]=ptr[i]=(idata) 10;
    }
    if  ( type==4 ) {
     bptr[i]=ptr[i]=(idata) i;
    }
    if  ( type>=5 ) {
     bptr[i]=ptr[i]=(idata) n-i+1;
    }
   }
     if (DEBUG) { 
        printf("%6d(%2d %2d)In  :",n,THRESI16,THRESI16);
        for(i=0;i<ELMNTS;i++) printf("%d ",(idata)bptr[i]); printf("..."); 
        for(i=n-ELMNTS;i<n;i++) printf("%d ",(idata)bptr[i]); printf("\n");
     }

     WRITE1;
     qsort(ptr,n,sizeof(idata),compare);
     WRITE2; 
     printf("(%2d,%2d,%9d,%1d,%1d,%1d):",THRESI16,TMERGEI,n,type,SPLITTER,runs);
     printf("    %18s generic: Elapsed time is: %10.8f\n","qsort",elapsed/runs);

     memcpy(obptr,ptr,n*sizeof(idata));

     if (DEBUG) { 
        printf("%6d(%2d %2d)Out :",n,THRESI16,THRESI16);
        for(i=0;i<ELMNTS;i++) printf("%d ",(idata)obptr[i]); printf("..."); 
        for(i=n-ELMNTS;i<n;i++) printf("%d ",(idata)obptr[i]); printf("\n");
     }


   for(k=0;k<N_SORT_ALGS;k++) {
    if (doalg[k]>=1) {
      GTYPESORT=gfunc[k];


      elapsed=0.0;
      for(j=0;j<runs;j++) {
           memcpy(ptr,bptr,n*sizeof(idata));
           if (DEBUG) { 
              printf("%6d(%2d %2d)In  :",n,THRESI16,THRESI16);
              for(i=0;i<ELMNTS;i++) printf("%d ",(idata)ptr[i]); printf("..."); 
              for(i=n-ELMNTS;i<n;i++) printf("%d ",(idata)ptr[i]); printf("\n");
           }

          if (k==12){
           avg_tstart();
               ignu01(ptr,n,sizeof(idata),compare);
           avg_tend(); elapsed += avg_treturn();
           }
           else {
           avg_tstart();
               GTYPESORT(ptr,n);
           avg_tend(); elapsed += avg_treturn();
           }

           if (DEBUG) {
              printf("%6d(%2d %2d)Out :",n,THRESI16,THRESI16);
              for(i=0;i<ELMNTS;i++) printf("%d ",(idata)ptr[i]); printf("...");
              for(i=n-ELMNTS;i<n;i++) printf("%d ",(idata)ptr[i]); printf("\n");
           }
      }

      for(i=0;i<n-1;i++) { 
             if (ptr[i]!=obptr[i]) {printf("i=%d ERR",i); break;}
      }

     printf("(%2d,%2d,%9d,%1d,%1d,%1d):",THRESI16,TMERGEI,n,type,SPLITTER,runs);


      if (memcmp(ptr,obptr,n*sizeof(idata))  != 0)
        printf("ERR %18s generic: Elapsed time is: %10.8f\n",labels[k],elapsed/runs);
      else
        printf("    %18s generic: Elapsed time is: %10.8f\n",labels[k],elapsed/runs);
    }
   }

   free((void *) obptr);
   free((void *) bptr);
   free((void *) ptr);
   return(0);
/*
   Free((void *) bptr,n*sizeof(idata),"bptr");
   Free((void *) ptr,n*sizeof(idata),"ptr");
*/
}
